# -*- coding: utf-8 -*-

"""Top-level package for Noisy Duck."""

__author__ = """Nathan Wukie"""
__email__ = 'nathan.wukie@gmail.com'
__version__ = '0.2.0'

from . import annulus
from . import filter
from .io import write_decomposition




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Testing decomposition results for cylindrical annulus.

    Fixtures:
        uniform_state:      returns dictionary containing the state for a uniform axial 
                            mean flow.
        uniform_numerical:  returns eigenvalues and eigenvectors from the numerical 
                            decomposition of the uniform axial mean flow.
        uniform_analytical: returns eigenvalues and eigenvectors from the analytical
                            decomposition of the uniform axial mean flow.

    The fixtures here essentially precompute results, then these results are
    used in tests for various purposes.

"""
from __future__ import division
import pytest
import numpy as np
import noisyduck as nd




@pytest.fixture
def uniform_state():
    """Return the mean state for uniform axial flow."""
    ri=0.25
    ro=1.0
    res=50
    r=np.linspace(ri,ro,res)
    return {'gam':1.4, 'rho':1.2, 'vr' :0., 'vt' :0., 'vz' :100., 'p':100000., 'omega':3000., 'm':2, 'r':r}




@pytest.fixture
def uniform_numerical(uniform_state):
    """Compute the numerical eigendecomposition of a uniform axial 
    mean flow that can be used by tests.
    """
    
    evals, evecs_l, evecs_r = nd.annulus.numerical.decomposition(uniform_state['omega'],
                                                                 uniform_state['m'],
                                                                 uniform_state['r'],
                                                                 uniform_state['rho'],
                                                                 uniform_state['vr'],
                                                                 uniform_state['vt'],
                                                                 uniform_state['vz'],
                                                                 uniform_state['p'],
                                                                 uniform_state['gam'],
                                                                 filter='acoustic',alpha=0.0000001,perturb_omega=False)
    return evals, evecs_l, evecs_r




@pytest.fixture
def uniform_analytical(uniform_state,uniform_numerical):
    """ Compute the analytical eigendecomposition of a uniform axial
    mean flow that can be used by tests.
    """
    gam  = uniform_state['gam']
    p    = uniform_state['p']
    rho  = uniform_state['rho']
    c    = np.sqrt(gam*p/rho)
    mach = uniform_state['vz']/c

    # Access the results of the numerical decomposition so we know how many
    # eigenvalues to find.
    evals_n, evecs_ln, evecs_rn = uniform_numerical
    n = len(evals_n)

    evals, evecs_r = nd.annulus.analytical.decomposition(uniform_state['omega']/c,
                                                         uniform_state['m'],
                                                         mach,
                                                         uniform_state['r'],
                                                         n)

    return evals, evecs_r





def test_uniformaxialflow_analytical_unique(uniform_analytical):
    """ Test the eigenvalues from the analytical decomposition are unique """
    evals, evecs_r = uniform_analytical
    assert len(evals) == len(set(evals))




def test_uniformaxialflow_analytical_matches_reference(uniform_analytical):
    """ Test the analytical eigenvalues match previously computed reference values. """

    evals_a, evecs_ra = uniform_analytical

    # Previously computed reference eigenvalues for the analytical
    # eigendecomposition of a uniform axial mean flow in an annular
    # duct. rho=1.2, vz=100, p=100000, omega=3000/c, m=2, ri/ro=0.25
    evals_ref = np.array([(11.888859931191348+0j),
                          (-6.263859931191352+0j),
                          (9.746593068700136+0j),
                          (-4.121593068700137+0j),
                          (2.8124999999999987+3.0004194877589363j),
                          (2.8124999999999987-3.0004194877589363j),
                          (2.812499999999999+10.162007091553026j),
                          (2.812499999999999-10.162007091553026j),
                          (2.8124999999999982+15.38697739666744j),
                          (2.8124999999999982-15.38697739666744j)])


    eval_matches = np.zeros(len(evals_a), dtype=bool)
    for i in range(len(evals_a)):
        for j in range(len(evals_ref)):
            if np.isclose(evals_a[i],evals_ref[j], atol=1.e-6):
                eval_matches[i]=True
                break

    # Assert that we found a matching reference eigenvalue for each
    # analytical eigenvalue.
    assert eval_matches.all()


def test_uniformaxialflow_numerical_matches_analytical(uniform_analytical,uniform_numerical):
    """ Test the annular cylindrical duct decomposition for the
    case of uniform axial mean flow.
    """
    # Unpack results from analytical and numerical decompositions
    evals_a, evecs_ra           = uniform_analytical
    evals_n, evecs_ln, evecs_rn = uniform_numerical

    # For each numerical eigenvalue, test that it matches closely
    # with one of the analytical eigenvalues.
    eval_matches = np.zeros(len(evals_n), dtype=bool)
    for i in range(len(evals_n)):
        # Search each entry in evals_a to try and find a close match
        for j in range(len(evals_a)):
            if np.isclose(evals_n[i],evals_a[j],atol=1.e-2): 
                eval_matches[i]=True 
                break

    # Assert that we found a matching analytical eigenvalue for each 
    # numerical eigenvalue
    assert eval_matches.all()











import numpy as np

# 15-point Dispersion Relation Preserving Filter
def drp15(f):
    """ 15-point DRP filter from "Computational Aeroacoustics", Tam. Strongly
    filters low-wavenumber components of a signal. End points are handled by
    creating an artificial reflection of the origin signal about the end points.
    In this way, the central filter has 'ghost' information past the end points
    that it needs to construct the filter.

    Args:
        f (np.array(float)): discrete values of f evaluated on a uniform grid.

    Returns:
        f_tilde (np.array(float)): filtered function
    """

    # 15-point DRP filter coefficients from "Computational Aeroacoustics", Tam
    d0  =  0.2042241813072920
    dp1 = -0.1799016298200503
    dm1 = -0.1799016298200503
    dp2 =  0.1224349282118140
    dm2 =  0.1224349282118140
    dp3 = -6.3456279827554890e-2
    dm3 = -6.3456279827554890e-2
    dp4 =  2.4341225689340974e-2
    dm4 =  2.4341225689340974e-2
    dp5 = -6.5519987489327603e-3
    dm5 = -6.5519987489327603e-3
    dp6 =  1.1117554451990776e-3
    dm6 =  1.1117554451990776e-3
    dp7 = -9.0091603462069583e-5
    dm7 = -9.0091603462069583e-5
    coeffs = [dm7, dm6, dm5, dm4, dm3, dm2, dm1, d0, dp1, dp2, dp3, dp4, dp5, dp6, dp7]

    # Create storage for result
    f_tilde = np.copy(f)

    # Duplicate waveform to handle end points
    two = np.append(np.flipud(f),f)
    three = np.append(two,np.flipud(f))

    # Apply filter
    for j in range(len(f)):
        dat = three[(len(f)+j-7):(len(f)+1+j+7)]
        f_tilde[j] = np.sum(dat*coeffs)

    return f_tilde




def physical(eigenvalues,ul,ur,r,alpha_cutoff=0.00001,filters='acoustic'):
    """ Procedure for filtering/sorting eigenvectors into physical categories.
    Generally, we are trying to determine if a given eigenvector is a convected
    wave, an upstream/downstream traveling acoustic wave, or a spurious mode.

    References:
    [1] Maldonado, A.L.P., Astley, R. J., Coupland, J., Gabard, G., and Sutliff, D., 
        "Sound Propagation in Lined Annular Ducts with Mean Swirling Flow" 21st AIAA/CEAS 
        Aeroacoustics Conference, June 2015. AIAA 2015-2522.

    Args:
        eigenvalues (complex): Array of eigenvalues.
        ul (complex): Array of left eigenvectors; one for each eigenvalue.  [len(r)*nfields,len(eigenvalues)].
        ur (complex): Array of right eigenvectors; one for each eigenvalue. [len(r)*nfields,len(eigenvalues)].
        r (float): Array of radial coordinate locations where the eigenvectors have been evaluated at.
        alpha_cutoff (float, optional): A cutoff criteria for filtering. Higher alpha will include higher wavenumber modes.
        filters (string, optional): Select how to filter the incoming eigenvalue/eigenvector pairs.

    Returns:
        (eigenvalues, eigenvectors): a tuple containing an array of filtered eigenvalues, and their corresponding eigenvectors.
    """
    # Separate eigenvectors into primitive variables
    res = len(r)
    rho_eigenvectors = np.copy(ur[0*res:1*res,:])
    u_eigenvectors   = np.copy(ur[1*res:2*res,:])
    v_eigenvectors   = np.copy(ur[2*res:3*res,:])
    w_eigenvectors   = np.copy(ur[3*res:4*res,:])
    p_eigenvectors   = np.copy(ur[4*res:5*res,:])

    if (filters == 'acoustic'):
        # Here, we work with the pressure component of the eigenvectors
        # Create separate copy so we can compare before/after modes
        #
        # We pick some arbitrary point in the eigenvector and compute the angle
        #   phi = atan2(Im{p'},Re{p'})
        #
        # Now, rotate the original vector so it is strictly real-valued so that
        # the sorting algorithm can just look at the real part of the eigenvector 
        # and doesn't accidentally see something that is zero, but just happened
        # to be at the wrong phase so all the energy was in the imaginary part.
        phi = np.arctan2(p_eigenvectors[5,:].imag, p_eigenvectors[5,:].real)
        for i in range(p_eigenvectors.shape[1]):
            p_eigenvectors[:,i] = p_eigenvectors[:,i]*np.exp(-1j*phi[i])


        # Just handling the real part now.
        pr_eigenvectors   = np.copy(np.real(p_eigenvectors))
        pr_eigenvectors_f = np.copy(pr_eigenvectors)

        # Call filter for each eigenvector
        for i in range(pr_eigenvectors.shape[1]):
            pr_eigenvectors_f[:,i] = drp15(pr_eigenvectors[:,i])

        # Select low-wavenumber modes based on the ratio of their vector norms
        amp   = np.zeros(pr_eigenvectors.shape[1])
        amp_f = np.zeros(pr_eigenvectors.shape[1])
        for i in range(pr_eigenvectors.shape[1]):
            for j in range(len(r)-1):
                amp[i]   = amp[i]   + 0.5*(r[j+1]-r[j])*(r[j]*pr_eigenvectors[  j,i]*pr_eigenvectors[  j,i]  +  r[j+1]*pr_eigenvectors[  j+1,i]*pr_eigenvectors[  j+1,i])
                amp_f[i] = amp_f[i] + 0.5*(r[j+1]-r[j])*(r[j]*pr_eigenvectors_f[j,i]*pr_eigenvectors_f[j,i]  +  r[j+1]*pr_eigenvectors_f[j+1,i]*pr_eigenvectors_f[j+1,i])

        # Evaluate wavenumber selection criteria: alpha
        alpha = np.zeros(len(amp))
        for i in range(len(amp)):
            if (amp[i] < np.finfo(float).eps):
                # Handle case where initial amplitude for pressure was zero. 
                # We aren't interested in those modes for acoustics so we set their metric to 
                # very large, effectively filtering them out.
                alpha[i] = np.finfo(float).max
            else:
                alpha[i] = amp_f[i]/amp[i]

        # Evaluate pressure magnitude selection criteria
        # Find 2*res eigenvectors with largest |p'|
        # np.argpartition just returns the indices
        for i in range(len(amp)):
            largest = np.argpartition(amp, -2*res)[-2*res:]


        # Flag eigenvectors that satisfy criteria: 
        #   1: Wavenumber filter criteria: alpha < alpha_cutoff.
        #   2: Pressure magnitude criteria, only 2*res largest are kept.
        keep = []
        for i in range(len(alpha)):
            if (np.abs(alpha[i]) < alpha_cutoff) and (i in largest):
                # Store as structured array so we can sort the entire system by alpha
                tmp = np.array([(i,alpha[i])],dtype=[('index',np.int),('alpha',np.float)])
                if (len(keep) == 0):
                    keep = tmp
                else:
                    keep = np.append(keep,tmp, axis=0)
        
        # Sort eigenmodes and indices being kept by 'alpha'
        keep = np.sort(keep,order='alpha')

        # Collect eigenvalues/eigenvectors flagged to pass the filter
        nmodes = keep.shape[0]
        eigenvalues_f = np.zeros([nmodes],       dtype=np.complex)
        ul_f          = np.zeros([5*res,nmodes], dtype=np.complex)
        ur_f          = np.zeros([5*res,nmodes], dtype=np.complex)
        for i in range(nmodes):
            eigenvalues_f[i] = np.copy(eigenvalues[keep['index'][i]])
            ur_f[:,i]        = np.copy(ur[:,keep['index'][i]])
            ul_f[:,i]        = np.copy(ul[:,keep['index'][i]])

        return eigenvalues_f, ul_f, ur_f
        




# -*- coding: utf-8 -*-
r""" 
Numerical
---------

This module provides a functionality for computing the numerical eigenvalue/eigenvector
decomposition of a uniform axial flow in an annular cylindrical duct. The decomposition
is based on a normal mode analysis of the three-dimensional linearized Euler equations,
which yields an eigensystem that is discretized and solved numerically.


Theory:
~~~~~~~



Filtering:
~~~~~~~~~~



Example:
~~~~~~~~

::
    
    eigenvalues, eigenvectors_l, eigenvectors_r = noisyduck.annulus.numerical.decomposition(omega,m,r,rho,vr,vt,vz,p,gam,filter='acoustic')

"""
import numpy as np
import scipy
import noisyduck.filter






def decomposition(omega,m,r,rho,vr,vt,vz,p,gam,filter='None',alpha=0.00001,equation='general',perturb_omega=True):
    r""" Compute the numerical eigen-decomposition of the three-dimensional linearized
    Euler equations for a cylindrical annulus.

    Args:
        omega (float): temporal frequency.
        m (int): circumferential wavenumber.
        r (float): array of equally-spaced radius locations for the discretization, including end points.
        rho (float): mean density.
        vr (float): mean radial velocity.
        vt (float): mean tangential velocity.
        vz (float): mean axial velocity.
        p (float): mean pressure.
        gam (float): ratio of specific heats.
        filter (string, optional): Optional filter for eigenmodes. values = ['None', 'acoustic']
        alpha (float, optional): Criteria governing filtering acoustic modes. 
        equation (string, optional): Select from of governing equation for the decomposition. values = ['general', 'radial equilibrium']
        perturb_omega (bool): If true, small imaginary part is added to the temporal frequency. Can help in determining direction of propagation.

    Returns:
        (eigenvalues, left_eigenvectors, right_eigenvectors): a tuple containing an array of eigenvalues, an array 
        of left eigenvectors evaluated at radial locations, an array of right eigenvectors evaluated at radial locations.

    Note:
        The eigenvectors being returned include each field :math:`[\rho,v_r,v_t,v_z,p]`. The primitive
        variables can be extracted into their own eigenvectors by copying out those entries from 
        the returned eigenvectors as:

        ::

            res = len(r)
            rho_eigenvectors = eigenvectors[0*res:1*res,:]
            vr_eigenvectors  = eigenvectors[1*res:2*res,:]
            vt_eigenvectors  = eigenvectors[2*res:3*res,:]
            vz_eigenvectors  = eigenvectors[3*res:4*res,:]
            p_eigenvectors   = eigenvectors[4*res:5*res,:]

    """
    res = len(r)
    ri = np.min(r)
    ro = np.max(r)


    # Construct eigensystem
    if (equation == 'general'):
        M, N = construct_numerical_eigensystem_general(omega,m,r,rho,vr,vt,vz,p,gam,perturb_omega)
    elif (equation == 'radial equilibrium'):
        M, N = construct_numerical_eigensystem_radial_equilibrium(omega,m,r,rho,vr,vt,vz,p,gam,perturb_omega)


    # Solve Standard Eigenvalue Problem for complex, nonhermitian system
    #evals, evecs_l, evecs_r = scipy.linalg.eig(np.linalg.inv(N)@M,left=True,right=True,overwrite_a=True,overwrite_b=True)
    evals, evecs_l, evecs_r = scipy.linalg.eig(np.matmul(np.linalg.inv(N),M),left=True,right=True,overwrite_a=True,overwrite_b=True)


    # Add radial velocity end points back where they were removed due to boundary conditions
    evecs_r = np.insert(evecs_r, [res]    , [0.] ,axis=0)
    evecs_r = np.insert(evecs_r, [2*res-1], [0.] ,axis=0)
    evecs_l = np.insert(evecs_l, [res]    , [0.] ,axis=0)
    evecs_l = np.insert(evecs_l, [2*res-1], [0.] ,axis=0)

    # Filtering
    if (filter == 'acoustic'):
        evals, evecs_l, evecs_r = noisyduck.filter.physical(evals,evecs_l,evecs_r,r,alpha_cutoff=alpha,filters=filter)

    # Return conventional definition of the left eigenvector
    evecs_l = np.copy(evecs_l.conj())

    return evals, evecs_l, evecs_r




def construct_numerical_eigensystem_general(omega,m,r,rho,vr,vt,vz,p,gam,perturb_omega=True):
    r""" Constructs the numerical representation of the eigenvalue problem associated
    with the three-dimensional linearized euler equations subjected to a normal mode
    analysis.

    NOTE: If perturb_omega=True, a small imaginary part is added to the temporal 
    frequency to facilitate determining the propagation direction of eigenmodes based 
    on the sign of the imaginary part of their eigenvalue. 
    That is: :math:`\omega = \omega - 10^{-5}\omega j`. See Moinier and Giles[2].

    [1] Kousen, K. A., "Eigenmodes of Ducted Flows With Radially-Dependent Axial 
        and Swirl Velocity Components", NASA/CR 1999-208881, March 1999.
    [2] Moinier, P., and Giles, M. B., "Eigenmode Analysis for Turbomachinery 
        Applications", Journal of Propulsion and Power, Vol. 21, No. 6, 
        November-December 2005.

    Args:
        omega (float): temporal frequency.
        m (int): circumferential wavenumber.
        r (float): array of equally-spaced radius locations for the discretization, including end points.
        rho (float): mean density.
        vr (float): mean radial velocity.
        vt (float): mean tangential velocity.
        vz (float): mean axial velocity.
        p (float): mean pressure.
        gam (float): ratio of specific heats.
        perturb_omega (bool): If true, small imaginary part is added to the temporal frequency. Can help in determining direction of propagation.

    Returns:
        (M, N): left-hand side of generalized eigenvalue problem, right-hand side of generalized eigenvalue problem.
    """

    # Define real/imag parts for temporal frequency
    romega = omega
    if (perturb_omega):
        iomega = -10.e-5*romega
    else:
        iomega = 0.


    # Define geometry and discretization
    res = len(r)
    ri = np.min(r)
    ro = np.max(r)
    dr = (ro-ri)/(res-1)
    nfields = 5
    dof = res*nfields


    # Check if input mean quantities are scalar. 
    # If so, expand them to a vector
    if (type(rho) is float):
       rho = np.full((res),rho)
    if (type(vr) is float):
       vr = np.full((res),vr)
    if (type(vt) is float):
       vt = np.full((res),vt)
    if (type(vz) is float):
       vz = np.full((res),vz)
    if (type(p) is float):
       p = np.full((res),p)



    # Allocate storage
    M  = np.zeros([dof,dof], dtype=np.complex)
    N  = np.zeros([dof,dof], dtype=np.complex)
    
    # Submatrices for discretization
    stencil     = np.zeros([res,res])
    identity    = np.zeros([res,res])
    ridentity   = np.zeros([res,res])
    zero_matrix = np.zeros([res,res])

    # Construct fourth-order finite difference stencil
    #
    # Stencil operations
    #
    #   stencil*f                     =>  d(f u')/dr   => returns matrix operator, L corresponding to d(f*u')/dr, where Lu' = d(f*u')/dr
    #
    #   #stencil@f                    =>  d(f)/dr      => returns vector of evaluated derivatives, d(f)/dr
    #   np.matmul(stencil,f)          =>  d(f)/dr      => returns vector of evaluated derivatives, d(f)/dr
    #
    #   #(identity*f)@stencil         => f d(u')/dr    => returns matrix operator, L corresponding to (f d()/dr), where Lu' = f*d(u')/dr
    #   np.matmul(identity*f,stencil) => f d(u')/dr    => returns matrix operator, L corresponding to (f d()/dr), where Lu' = f*d(u')/dr
    #
    stencil[0,0:5] = [-25., 48., -36, 16., -3.]
    stencil[1,0:5] = [-3., -10., 18., -6., 1.]
    stencil[res-2, res-5:res] = [-1., 6., -18., 10., 3.]
    stencil[res-1, res-5:res] = [3., -16., 36., -48., 25.]
    for i in range(2,res-2):
        stencil[i,i-2] =  1.
        stencil[i,i-1] = -8.
        stencil[i,i+1] =  8.
        stencil[i,i+2] = -1.
    # Scale by dr
    stencil = (1./(12.*dr))*stencil
    
    
    # Construct identity matrix for source terms
    for i in range(res):
        identity[i,i] = 1.
    
    
    # Construct identity scaled by 1/r
    for i in range(res):
        ridentity[i,i] = 1./r[i]
    

    #
    # Compute radial derivatives
    #
    #drho_dr = stencil@rho
    #dvr_dr  = stencil@vr
    #dvt_dr  = stencil@vt
    #dvz_dr  = stencil@vz
    #dp_dr   = stencil@p
    drho_dr = np.matmul(stencil,rho)
    dvr_dr  = np.matmul(stencil,vr)
    dvt_dr  = np.matmul(stencil,vt)
    dvz_dr  = np.matmul(stencil,vz)
    dp_dr   = np.matmul(stencil,p)



    # Block 1,1
    irow = 1
    icol = 1
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega                # temporal source(real)
                                                                        -    identity*iomega                # temporal source(imag)
                                                                        #+    (identity*vr)@stencil         # bar{B} radial
                                                                        +    np.matmul(identity*vr,stencil) # bar{B} radial
                                                                        +    identity*dvr_dr                # tilde{B}
                                                                        + 1j*ridentity*float(m)*vt          # bar{C} circumferential source
                                                                        +    ridentity*vr )                 # bar{D} 

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*vz )                  # axial source

    # Block 2,2
    irow = 2
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega                # temporal source(real)
                                                                        -    identity*iomega                # temporal source(imag)
                                                                        #+    (identity*vr)@stencil         # bar{B} radial derivative
                                                                        +    np.matmul(identity*vr,stencil) # bar{B} radial derivative
                                                                        +    identity*dvr_dr                # tilde{B}
                                                                        + 1j*ridentity*float(m)*vt )        # bar{C} circumferential source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*vz )                  # bar{A} axial source


    # Block 3,3
    irow = 3
    icol = 3
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega                # temporal source(real)
                                                                        -    identity*iomega                # temporal source(imag)
                                                                        #+    (identity*vr)@stencil         # bar{B} radial derivative
                                                                        +    np.matmul(identity*vr,stencil) # bar{B} radial derivative
                                                                        + 1j*ridentity*float(m)*vt          # bar{C} circumferential source
                                                                        +    ridentity*vr )                 # bar{D} equation/coordinate source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*vz )                  # bar{A} axial source



    # Block 4,4
    irow = 4
    icol = 4
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega                # temporal source(real)
                                                                        -    identity*iomega                # temporal source(imag)
                                                                        #+    (identity*vr)@stencil         # bar{B} radial derivative
                                                                        +    np.matmul(identity*vr,stencil) # bar{B} radial derivative
                                                                        + 1j*ridentity*float(m)*vt )        # bar{C} circumferential source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*vz )                  # bar{A} axial source


    # Block 5,5
    irow = 5
    icol = 5
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega                # temporal source(real)
                                                                        -    identity*iomega                # temporal source(imag)
                                                                        #+    (identity*vr)@stencil         # bar{B} radial derivative
                                                                        +    np.matmul(identity*vr,stencil) # bar{B} radial derivative
                                                                        +     identity*dvr_dr*gam           # tilde{B}
                                                                        + 1j*ridentity*float(m)*vt          # bar{C} circumferential source
                                                                        +    ridentity*vr*gam )             # bar{D} equation/coordinate source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*vz )                  # bar{A} axial source


    # Block 2,1
    irow = 2
    icol = 1
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        - ridentity*vt*vt/rho               # bar{D} equation/coordinate source
                                                                        + identity*(vr/rho)*dvr_dr )        # tilde{B}

    # Block 3,1
    irow = 3
    icol = 1
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + ridentity*vr*vt/rho               # bar{D} equation/coordinate source
                                                                        + identity*(vr/rho)*dvt_dr )        # tilde{B}

    # Block 4,1
    irow = 4
    icol = 1
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + identity*(vr/rho)*dvz_dr )        # tilde{B}

    # Block 1,2
    irow = 1
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        #+ (identity*rho)@stencil           # bar{B} radial derivative
                                                                        + np.matmul(identity*rho,stencil)   # bar{B} radial derivative
                                                                        + ridentity*rho                     # bar{D} equation/coordinate source
                                                                        + identity*drho_dr )                # tilde{B}

    # Block 3,2
    irow = 3
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + ridentity*vt                      # bar{D} equation/coordinate source
                                                                        + identity*dvt_dr )                 # tilde{B}

    # Block 4,2
    irow = 4
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + identity*dvz_dr )                 # tilde{B}

    # Block 5,2
    irow = 5
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        #+ (identity*(p*gam))@stencil           # bar{B} radial derivative
                                                                        + np.matmul(identity*(p*gam),stencil)   # bar{B} radial derivative
                                                                        + ridentity*p*gam                       # bar{D} equation/coordinate source
                                                                        +  identity*dp_dr )                     # tilde{B}

    # Block 1,3
    irow = 1
    icol = 3
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*ridentity*rho*float(m) )           # bar{C} circumferential source

    # Block 2,3
    irow = 2
    icol = 3
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        - ridentity*2.*vt )                     # bar{D} equation/coordinate source

    # Block 5,3
    irow = 5
    icol = 3
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*ridentity*gam*p*float(m) )         # bar{C} circumferential source

    # Block 1,4
    irow = 1
    icol = 4
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*rho )                     # bar{A} axial source

    # Block 5,4
    irow = 5
    icol = 4
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*gam*p )                   # bar{A} axial source

    # Block 2,5
    irow = 2
    icol = 5
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        #+ (identity*(1./rho))@stencil )            # bar{B} radial derivative
                                                                        + np.matmul(identity*(1./rho),stencil) )    # bar{B} radial derivative

    # Block 3,5
    irow = 3
    icol = 5
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*ridentity*float(m)/rho )       # bar{C} circumferential source

    # Block 4,5
    irow = 4
    icol = 5
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity/rho )                 # bar{A} axial source



    # Remove rows/columns due to solid wall boundary condition: no velocity normal to boundary, assumes 
    # here wall normal vector is aligned with radial coordinate.
    M = np.delete(M, (res,2*res-1), axis=0)
    M = np.delete(M, (res,2*res-1), axis=1)
    
    N = np.delete(N, (res,2*res-1), axis=0)
    N = np.delete(N, (res,2*res-1), axis=1)
    
    
    # Move N to right-hand side
    N = np.copy(-N)


    return M, N









def construct_numerical_eigensystem_radial_equilibrium(omega,m,r,rho,vr,vt,vz,p,gam,perturb_omega=True):
    """ Constructs the numerical representation of the eigenvalue problem associated
    with the three-dimensional linearized euler equations under the assumption of 
    radial equilibrium subjected to a normal mode analysis.

    NOTE: If perturb_omega=True, a small imaginary part is added to the temporal 
    frequency to facilitate determining the propagation direction of eigenmodes based 
    on the sign of the imaginary part of their eigenvalue. 
    That is: :math:`\omega = \omega - 10^{-5}\omega j`. See Moinier and Giles[2].

    The equation set used for this decomposition is consistent with that presented by
    Sharma et. al[1]. Even for radial equilibrium flows, this equation set is missing 
    a dvt_dr term in the tangential velocity equation and also a drho_dr term in the
    radial velocity equation.

    References:
    [1] Sharma, A., Richards, S. K., Wood, T. H., Shieh, C., "Numerical Prediction 
        of Exhaust Fan-Tone Noise from High-Bypass Aircraft Engines", AIAA Journal, 
        Vol. 47, No. 12, December 2009.
    [2] Moinier, P., and Giles, M. B., "Eigenmode Analysis for Turbomachinery 
        Applications", Journal of Propulsion and Power, Vol. 21, No. 6, 
        November-December 2005.
    [3] Kousen, K. A., "Eigenmodes of Ducted Flows With Radially-Dependent Axial 
        and Swirl Velocity Components", NASA/CR 1999-208881, March 1999.

    Args:
        omega (float): temporal frequency.
        m (int): circumferential wavenumber.
        r (float): array of equally-spaced radius locations for the discretization, including end points.
        rho (float): mean density.
        vr (float): mean radial velocity.
        vt (float): mean tangential velocity.
        vz (float): mean axial velocity.
        p (float): mean pressure.
        gam (float): ratio of specific heats.

    Returns:
        (M, N): left-hand side of generalized eigenvalue problem, right-hand side of generalized eigenvalue problem.
    """

    # Define real/imag parts for temporal frequency
    romega = omega
    if (perturb_omega):
        iomega = -10.e-5*romega
    else:
        iomega = 0.

    # Define geometry and discretization
    res = len(r)
    ri = np.min(r)
    ro = np.max(r)
    dr = (ro-ri)/(res-1)
    nfields = 5
    dof = res*nfields


    # Check if input mean quantities are scalar. 
    # If so, expand them to a vector
    if (type(rho) is float):
       rho = np.full((res),rho)
    if (type(vr) is float):
       vr = np.full((res),vr)
    if (type(vt) is float):
       vt = np.full((res),vt)
    if (type(vz) is float):
       vz = np.full((res),vz)
    if (type(p) is float):
       p = np.full((res),p)






    # Allocate storage
    M  = np.zeros([dof,dof], dtype=np.complex)
    N  = np.zeros([dof,dof], dtype=np.complex)
    
    # Submatrices for discretization
    stencil     = np.zeros([res,res])
    identity    = np.zeros([res,res])
    ridentity   = np.zeros([res,res])
    zero_matrix = np.zeros([res,res])

    # Construct fourth-order finite difference stencil
    stencil[0,0:5] = [-25., 48., -36, 16., -3.]
    stencil[1,0:5] = [-3., -10., 18., -6., 1.]
    stencil[res-2, res-5:res] = [-1., 6., -18., 10., 3.]
    stencil[res-1, res-5:res] = [3., -16., 36., -48., 25.]
    for i in range(2,res-2):
        stencil[i,i-2] =  1.
        stencil[i,i-1] = -8.
        stencil[i,i+1] =  8.
        stencil[i,i+2] = -1.
    # Scale by dr
    stencil = (1./(12.*dr))*stencil
    
    
    # Construct identity matrix for source terms
    for i in range(res):
        identity[i,i] = 1.
    
    
    # Construct identity scaled by 1/r
    for i in range(res):
        ridentity[i,i] = 1./r[i]
    

    #
    # Compute radial derivatives
    #
    #drho_dr = stencil@rho
    #dvr_dr  = stencil@vr
    #dvt_dr  = stencil@vt
    #dvz_dr  = stencil@vz
    #dp_dr   = stencil@p
    drho_dr = np.matmul(stencil,rho)
    dvr_dr  = np.matmul(stencil,vr)
    dvt_dr  = np.matmul(stencil,vt)
    dvz_dr  = np.matmul(stencil,vz)
    dp_dr   = np.matmul(stencil,p)





    # Block 1,1
    irow = 1
    icol = 1
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega         # temporal source(real)
                                                                        -    identity*iomega         # temporal source(imag)
                                                                        +    stencil*vr              # radial derivative
                                                                        +    ridentity*vr            # radial source
                                                                        + 1j*ridentity*float(m)*vt ) # circumferential source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*vz )           # axial source

    # Block 2,2
    irow = 2
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega        # temporal source(real)
                                                                        -    identity*iomega        # temporal source(imag)
                                                                        +    stencil*vr              # radial derivative
                                                                        +    ridentity*vr            # radial source
                                                                        + 1j*ridentity*float(m)*vt ) # circumferential source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*vz )           # axial source


    # Block 3,3
    irow = 3
    icol = 3
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega        # temporal source(real)
                                                                        -    identity*iomega        # temporal source(imag)
                                                                        +    stencil*vr              # radial derivative
                                                                        +    ridentity*vr            # radial source
                                                                        + 1j*ridentity*float(m)*vt   # circumferential source
                                                                        +    ridentity*vr )          # equation/coordinate source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*vz )           # axial source



    # Block 4,4
    irow = 4
    icol = 4
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega        # temporal source(real)
                                                                        -    identity*iomega        # temporal source(imag)
                                                                        +    stencil*vr              # radial derivative
                                                                        +    ridentity*vr            # radial source
                                                                        + 1j*ridentity*float(m)*vt ) # circumferential source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*vz )           # axial source


    # Block 5,5
    irow = 5
    icol = 5
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega        # temporal source(real)
                                                                        -    identity*iomega        # temporal source(imag)
                                                                        +    stencil*vr              # radial derivative
                                                                        +    ridentity*vr            # radial source
                                                                        + 1j*ridentity*float(m)*vt   # circumferential source
                                                                        +    ridentity*(gam-1.)*vr ) # equation/coordinate source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*vz )           # axial source


    # Block 2,1
    irow = 2
    icol = 1
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        - ridentity*vt*vt/rho )   # equation/coordinate source

    # Block 3,1
    irow = 3
    icol = 1
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + ridentity*vr*vt/rho )   # equation/coordinate source


    # Block 1,2
    irow = 1
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + stencil*rho           # radial derivative
                                                                        + ridentity*rho )       # radial source

    # Block 3,2
    irow = 3
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + ridentity*vt )         # equation/coordinate source

    # Block 5,2
    irow = 5
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + stencil*gam*p                   # radial derivative
                                                                        + ridentity*gam*p                 # radial source
                                                                        - ridentity*(gam-1.)*rho*vt*vt )  # equation/coordinate source

    # Block 1,3
    irow = 1
    icol = 3
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*ridentity*rho*float(m) )   # circumferential source

    # Block 2,3
    irow = 2
    icol = 3
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        - ridentity*2.*vt )              # equation/coordinate source

    # Block 5,3
    irow = 5
    icol = 3
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*ridentity*gam*p*float(m)       # circumferential source
                                                                        +    ridentity*(gam-1.)*rho*vr*vt )   # equation/coordinate source

    # Block 1,4
    irow = 1
    icol = 4
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*rho )                # axial source

    # Block 5,4
    irow = 5
    icol = 4
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*gam*p )              # axial source

    # Block 2,5
    irow = 2
    icol = 5
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + stencil/rho                   # radial derivative
                                                                        + ridentity/rho                 # radial source
                                                                        #-  identity*drho_dr/(rho*rho)   # TEST SOURCE
                                                                        - ridentity/rho )               # equation/coordinate source
    # Block 3,5
    irow = 3
    icol = 5
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*ridentity*float(m)/rho )   # circumferential source

    # Block 4,5
    irow = 4
    icol = 5
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity/rho )             # axial source



    # Remove rows/columns due to solid wall boundary condition: no velocity normal to boundary, assumes 
    # here wall normal vector is aligned with radial coordinate.
    M = np.delete(M, (res,2*res-1), axis=0)
    M = np.delete(M, (res,2*res-1), axis=1)
    
    N = np.delete(N, (res,2*res-1), axis=0)
    N = np.delete(N, (res,2*res-1), axis=1)
    
    
    # Move N to right-hand side
    N = np.copy(-N)


    return M, N








import numpy as np


def write_decomposition(eigenvalues, eigenvectors_l, eigenvectors_r, r, filename):
    """ Write the result of a numerical eigen-decomposition to file.

    Args:
        eigenvalues (complex): array of eigenvalues of the decomposition.
        eigenvectors_l (np.array([float,float])): rank-2 array of left eigenvectors.  [len(r)*nfields,len(nvectors)]
        eigenvectors_r (np.array([float,float])): rank-2 array of right eigenvectors. [len(r)*nfields,len(nvectors)].
        r (float): array of radial coordinates from the discretization that the eigenvectors were evaluated at.
        filename (string): filename where data will be written to. If it exists, it is overwritten.

    """
    # Open new file
    handle = open(filename,'w')

    handle.write('&sizes\n')
    handle.write('  nr       = '+str(len(r))+'\n')
    handle.write('  nvectors = '+str(len(eigenvalues))+'\n')
    handle.write('/\n')
    
    handle.write('&eigendecomposition\n')
    handle.write('  k = ')
    for i in range(len(eigenvalues)):
        handle.write('('+str(eigenvalues[i].real)+','+str(eigenvalues[i].imag)+')')
        if i == (len(eigenvalues)-1):
            handle.write('\n')
        else:
            handle.write(', ')
#    handle.write('/\n')


    handle.write('  r = ')
    for i in range(len(r)):
        handle.write(str(r[i]))
        if i == (len(r)-1):
            handle.write('  \n')
        else:
            handle.write(', ')


    # write right eigenvectors
    for vec in range(eigenvectors_r.shape[1]):
        handle.write('  A(1:'+str(eigenvectors_r.shape[0])+','+str(vec+1)+') = ')
        for val in range(eigenvectors_r.shape[0]):
            handle.write('('+str(eigenvectors_r[val,vec].real)+','+str(eigenvectors_r[val,vec].imag)+')')
            if val == eigenvectors_r.shape[0]-1:
                handle.write('\n')
            else:
                handle.write(', ')

    # write left eigenvectors
    for vec in range(eigenvectors_l.shape[1]):
        handle.write('  B(1:'+str(eigenvectors_l.shape[0])+','+str(vec+1)+') = ')
        for val in range(eigenvectors_l.shape[0]):
            handle.write('('+str(eigenvectors_l[val,vec].real)+','+str(eigenvectors_l[val,vec].imag)+')')
            if val == eigenvectors_l.shape[0]-1:
                handle.write('\n')
            else:
                handle.write(', ')
    handle.write('/\n')


    handle.close()





# -*- coding: utf-8 -*-
r""" 
Analytical
----------

This module provides a functionality for computing the eigenvalue/eigenvector
decomposition of a uniform axial flow in an annular cylindrical duct. The decomposition
is based on an analytical solution of the convected wave equation for pressure, 
yielding the acoustic part of the eigen decomposition. The eigenvectors from this 
decomposition correspond specifically to the acoustic pressure disturbance.


Theory:
~~~~~~~

The analysis here follows that of Moinier and Giles, "Eigenmode Analysis for 
Turbomachinery Applications", Journal of Propulsion and Power, Vol. 21, No. 6, 2005.

For a uniform axial mean, linear pressure perturbations satisfy the convected wave equation

.. math::

    \left( \frac{\partial}{\partial t} + M \frac{\partial}{\partial z} \right)^2 p = 
    \nabla^2 p \quad\quad\quad \lambda < r < 1


where :math:`M` is the Mach number of the axial mean flow and :math:`r` has been normalized
by the outer duct radius. Boundary conditions for a hard-walled duct imply zero radial 
velocity, which is imposed using the condition


.. math::

    \frac{\partial p}{\partial r} = 0   \quad\quad \text{at}    \quad\quad r=\lambda,1


Conducting a normal mode analysis of the governing equation involves assuming a solution
with a form 

.. math::

    p(r,\theta,z,t) = e^{j(\omega t + m \theta + k z)}P(r)

This leads to a Bessel equation

.. math::

    \frac{1}{r} \frac{d}{dr} \left(r \frac{dP}{dr} \right) + \left( \mu^2 - \frac{m^2}{r^2} 
    \right)P = 0    \quad\quad \lambda < r < 1

where :math:`\mu^2 = (\omega + M k)^2 - k^2`. The solution to the Bessel equation is

.. math::

    P(r) = a J_m(\mu r) + b Y_m(\mu r)

where :math:`J_m` and :math:`Y_m` are Bessel functions. Applying boundary conditions
to the general solution gives a set of two equations

.. math::

    \begin{bmatrix}
        J'_m(\mu \lambda)   &   Y'_m(\mu \lambda) \\
        J'_m(\mu)           &   Y'_m(\mu)
    \end{bmatrix}
    \begin{bmatrix}
        a \\ b
    \end{bmatrix}
    = 0

which has nontrivial solutions as long as the determinant is zero. Solving for the 
zeros of the determinant give :math:`\mu`, at which point the quadratic equation above
can be solved for the axial wavenumbers :math:`k`.

Example:
~~~~~~~~

::
    
    eigenvalues, eigenvectors = noisyduck.annulus.analytical.decomposition(omega,m,mach,r,n)

"""
import numpy as np
import scipy.special as sp
import scipy.optimize as op



def decomposition(omega,m,mach,r,n):
    """ This procedure computes the analytical eigen-decomposition of 
    the convected wave equation. The eigenvectors returned correspond specifically
    with acoustic pressure perturbations.

    Inner and outer radii are computed using min(r) and max(r), so it is important 
    that these end-points are included in the incoming array of radial locations.

    Args:
        omega (float): temporal wave number.
        m (int): circumferential wave number.
        mach (float): Mach number.
        r (float): array of radius stations, including rmin and rmax.
        n (int): number of eigenvalues/eigenvectors to compute.

    Returns:
        (eigenvalues, eigenvectors): a tuple containing an array of eigenvalues, and an array of eigenvectors evaluated at radial locations.

    """

    if (n % 2 != 0):
        print("WARNING: number of eigenvalues to find should be divisible by two. "
              "Because, for each zero found, there are two associated eigenvalues.")

    ri = np.min(r)
    ro = np.max(r)

    # Compute zeros for determinant of solution to convected wave equation
    # Then use zeros to compute the axial wavenumbers(eigenvalues)
    nzeros = int(n/2)
    zeros = compute_zeros(m,mach,ri,ro,nzeros)
    eigenvalues = compute_eigenvalues(omega,mach,zeros)

    #r = np.linspace(ri,ro,100)
    eigenvectors = np.zeros((len(r),n))
    for i in range(len(zeros)):
        eigenvectors[:,2*i  ] = compute_eigenvector(r,(ri/ro),m,zeros[i])
        eigenvectors[:,2*i+1] = compute_eigenvector(r,(ri/ro),m,zeros[i])

    return eigenvalues, eigenvectors




def eigensystem(b,m,ri,ro):
    r""" Computes the function associated with the eigensystem of the 
    convected wave equation. The location of the zeros for this function 
    correspond to the eigenvalues for the convected wave equation.

    The solution to the Bessel equation with boundary conditions applied
    yields a system of two linear equations.
    .. math::

        A x = 
        \begin{bmatrix}
            J'_m(\mu \lambda)   &   Y'_m(\mu \lambda) \\
            J'_m(\mu)           &   Y'_m(\mu)
        \end{bmatrix}
        \begin{bmatrix}
            x_1 \\ x_2
        \end{bmatrix}
        = 0

    This procedure evaluates the function
    .. math::

        det(A) = f(b) = J_m(b*ri)*Y_m(b*ro)  -  J_m(b*ro)*Y_m(b*ri) 

    So, this procedure can be passed to another routine such as numpy
    to find zeros.

    Args:
        b (float): coordinate.
        m (int): circumferential wave number.
        ri (float): inner radius of a circular annulus.
        ro (float): outer radius of a circular annulus.

    """
    f = sp.jvp(m,b*ri)*sp.yvp(m,b*ro) - sp.jvp(m,b*ro)*sp.yvp(m,b*ri)
    return f



def compute_zeros(m,mach,ri,ro,n):
    """ This procedure compute the zeros of the determinant for the convected
    wave equation. 
    
    A uniform set of nodes is created to search for sign changes
    in the value of the function. When a sign change is detected, it is known
    that a zero is close. The function is then passed to a bisection routine
    to find the location of the zero. 

    Args:
        m (int): circumferential wavenumber.
        mach (float): Mach number
        ri (float): inner radius of a circular annulus.
        ro (float): outer radius of a circular annulus.
        n (int): number of eigenvalues to compute.

    Returns:
        An array of the first 'n' eigenvalues for the system.

    """
    zero_loc = np.zeros(n)
    for rmode in range(n):
        if rmode == 0:
            #beta = np.linspace(0.5,200,2000)
            beta = np.linspace(0.000001,100,2000)
        else:
            beta = np.linspace(zero_loc[rmode-1]+0.1,200,20000)

        for i in range(len(beta)):
            f = eigensystem(beta[i],m,ri,ro)
            fnew = f
            bnew = beta[i]
            if i > 0 :
                #Test for sign change
                if np.sign(fnew) != np.sign(fold):
                    #Sign changed, so start bisection
                    zero_loc[rmode] = op.bisect(eigensystem,bold,bnew,(m,ri,ro))
                    break
                else:
                    fold = fnew
                    bold = bnew
            else:
                fold = fnew
                bold = bnew

    return zero_loc





def compute_eigenvalues(omega,mach,zeros):
    """ This procedure compute the analytical eigenvalues for the convected
    wave equation. A uniform set of nodes is created to search for sign changes
    in the value for the eigensystem. When a sign change is detected, it is known
    that a zero is close. The eigensystem is then passed to a bisection routine
    to find the location of the zero. The location corresponds to the eigenvalue
    for the system.

    Args:
        omega (float): temporal wavenumber.
        m (int): circumferential wavenumber.
        mach (float): Mach number.
        ri (float): inner radius of a circular annulus.
        ro (float): outer radius of a circular annulus.
        n (int): number of eigenvalues to compute.

    Returns:
        An array of the first 'n' eigenvalues for the system.

    """
    # Compute eigenvalues by solving: zeros^2 = (omega + M*k)^2 - k^2  for k
    k = np.zeros(2*len(zeros),dtype=np.complex)
    for i in range(len(zeros)):
        a = mach*mach - 1.
        b = 2.*omega*mach
        c = omega*omega - zeros[i]*zeros[i]
        roots = np.roots([a,b,c])
        k[2*i]   = roots[0]
        k[2*i+1] = roots[1]

    return k



def compute_eigenvector(r,sigma,m,zero):
    """ Return the eigenvector for the system.

    Args:
        r (np.array(float)): array of radial locations.
        sigma (float): ratio of inner to outer radius, ri/ro.
        m (int): circumferential wavenumber.
        zero (float): a zero of the determinant of the convected wave equation

    Returns:
        the eigenvector associated with the input 'm' and 'zero', evaluated at radial locations 
        defined by the input array 'r'. Length of the return array is len(r).

    """
    Q_mn = -sp.jvp(m,sigma*zero)/sp.yvp(m,sigma*zero)

    eigenvector = np.zeros(len(r))
    for irad in range(len(r)):
        eigenvector[irad] = sp.jv(m,zero*r[irad]) + Q_mn*sp.yv(m,zero*r[irad])

    # Normalize the eigenmode. Find abs of maximum value.
    ev_mag = np.absolute(eigenvector)
    ev_max = np.max(ev_mag)

    eigenvector = eigenvector/ev_max

    return eigenvector











